# -*- coding: UTF-8 -*-
# @Time    : 2017/5/16 11:22
# @Author  : HeYongbiao
# @email   : gzheyongbiao@corp.netease.com
import re
from collections import namedtuple
from datetime import datetime, timedelta
import time
import sys

_ranges = [
    (0, 59),
    (0, 23),
    (1, 31),
    (1, 12),
    (0, 6),
    (1970, 2099),
]
_attribute = [
    'minute',
    'hour',
    'day',
    'month',
    'isoweekday',
    'year'
]
_alternate = {
    3: {'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4, 'may': 5, 'jun': 6,
        'jul': 7, 'aug': 8, 'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12},
    4: {'sun': 0, 'mon': 1, 'tue': 2, 'wed': 3, 'thu': 4, 'fri': 5, 'sat': 6},
}
_aliases = {
    '@yearly': '0 0 1 1 *',
    '@annually': '0 0 1 1 *',
    '@monthly': '0 0 1 * *',
    '@weekly': '0 0 * * 0',
    '@daily': '0 0 * * *',
    '@hourly': '0 * * * *',
}

if sys.version_info >= (3, 0):
    _number_types = (int, float)
    xrange = range
else:
    _number_types = (int, long, float)

MINUTE = timedelta(minutes=1)
HOUR = timedelta(hours=1)
DAY = timedelta(days=1)
WEEK = timedelta(days=7)
MONTH = timedelta(days=28)
YEAR = timedelta(days=365)

WARN_CHANGE = object()


# find the next scheduled time
def _end_of_month(dt):
    ndt = dt + DAY
    while dt.month == ndt.month:
        dt += DAY
    return ndt.replace(day=1) - DAY


def _month_incr(dt, m):
    odt = dt
    dt += MONTH
    while dt.month == odt.month:
        dt += DAY
    # get to the first of next month, let the backtracking handle it
    dt = dt.replace(day=1)
    return dt - odt


def _year_incr(dt, m):
    # simple leapyear stuff works for 1970-2099 :)
    mod = dt.year % 4
    if mod == 0 and (dt.month, dt.day) < (2, 29):
        return YEAR + DAY
    if mod == 3 and (dt.month, dt.day) > (2, 29):
        return YEAR + DAY
    return YEAR


_increments = [
    lambda *a: MINUTE,
    lambda *a: HOUR,
    lambda *a: DAY,
    _month_incr,
    lambda *a: DAY,
    _year_incr,
    lambda dt, x: dt.replace(minute=0),
    lambda dt, x: dt.replace(hour=0),
    lambda dt, x: dt.replace(day=1) if x > DAY else dt,
    lambda dt, x: dt.replace(month=1) if x > DAY else dt,
    lambda dt, x: dt,
]


# find the previously scheduled time
def _day_decr(dt, m):
    if m.day.input != 'l':
        return -DAY
    odt = dt
    ndt = dt = dt - DAY
    while dt.month == ndt.month:
        dt -= DAY
    return dt - odt


def _month_decr(dt, m):
    odt = dt
    # get to the last day of last month, let the backtracking handle it
    dt = dt.replace(day=1) - DAY
    return dt - odt


def _year_decr(dt, m):
    # simple leapyear stuff works for 1970-2099 :)
    mod = dt.year % 4
    if mod == 0 and (dt.month, dt.day) > (2, 29):
        return -(YEAR + DAY)
    if mod == 1 and (dt.month, dt.day) < (2, 29):
        return -(YEAR + DAY)
    return -YEAR


def _day_decr_reset(dt, x):
    if x >= -DAY:
        return dt
    cur = dt.month
    while dt.month == cur:
        dt += DAY
    return dt - DAY


_decrements = [
    lambda *a: -MINUTE,
    lambda *a: -HOUR,
    _day_decr,
    _month_decr,
    lambda *a: -DAY,
    _year_decr,
    lambda dt, x: dt.replace(minute=59),
    lambda dt, x: dt.replace(hour=23),
    _day_decr_reset,
    lambda dt, x: dt.replace(month=12) if x < -DAY else dt,
    lambda dt, x: dt,
]

Matcher = namedtuple('Matcher', 'minute, hour, day, month, weekday, year')


def _assert(condition, message, *args):
    if not condition:
        raise ValueError(message % args)


class _Matcher(object):
    __slots__ = 'allowed', 'end', 'any', 'input', 'which', 'split'

    def __init__(self, which, entry):
        _assert(0 <= which <= 5,
                "improper number of cron entries specified")
        self.input = entry.lower()
        self.split = self.input.split(',')
        self.which = which
        self.allowed = set()
        self.end = None
        self.any = '*' in self.split or '?' in self.split

        for it in self.split:
            al, en = self._parse_crontab(which, it)
            if al is not None:
                self.allowed.update(al)
            self.end = en
        _assert(self.end is not None,
                "improper item specification: %r", entry.lower()
                )

    def __call__(self, v, dt):
        for i, x in enumerate(self.split):
            if x == 'l':
                if v == _end_of_month(dt).day:
                    return True

            elif x.startswith('l'):
                # We have to do this in here, otherwise we can end up, for
                # example, accepting *any* Friday instead of the *last* Friday.
                if dt.month == (dt + WEEK).month:
                    continue

                x = x[1:]
                if x.isdigit():
                    x = int(x) if x != '7' else 0
                    if v == x:
                        return True
                    continue

                start, end = map(int, x.partition('-')[::2])
                allowed = set(range(start, end + 1))
                if 7 in allowed:
                    allowed.add(0)
                if v in allowed:
                    return True

        return self.any or v in self.allowed

    def __lt__(self, other):
        if self.any:
            return self.end < other
        return all(item < other for item in self.allowed)

    def __gt__(self, other):
        if self.any:
            return _ranges[self.which][0] > other
        return all(item > other for item in self.allowed)

    def _parse_crontab(self, which, entry):
        """
        This parses a single crontab field and returns the data necessary for
        this matcher to accept the proper values.

        See the README for information about what is accepted.
        """

        # this handles day of week/month abbreviations
        def _fix(it):
            if which in _alternate and not it.isdigit():
                if it in _alternate[which]:
                    return _alternate[which][it]
            _assert(it.isdigit(),
                    "invalid range specifier: %r (%r)", it, entry)
            it = int(it, 10)
            _assert(_start <= it <= _end_limit,
                    "item value %r out of range [%r, %r]",
                    it, _start, _end_limit)
            return it

        # this handles individual items/ranges
        def _parse_piece(it):
            if '-' in it:
                start, end = map(_fix, it.split('-'))
                # Allow "sat-sun"
                if which == 4 and end == 0:
                    end = 7
            elif it == '*':
                start = _start
                end = _end
            else:
                start = _fix(it)
                end = _end
                if increment is None:
                    return set([start])

            _assert(_start <= start <= _end_limit,
                    "range start value %r out of range [%r, %r]",
                    start, _start, _end_limit)
            _assert(_start <= end <= _end_limit,
                    "range end value %r out of range [%r, %r]",
                    end, _start, _end_limit)
            _assert(start <= end,
                    "range start value %r > end value %r", start, end)
            return set(range(start, end + 1, increment or 1))

        _start, _end = _ranges[which]
        _end_limit = _end
        # wildcards
        if entry in ('*', '?'):
            if entry == '?':
                _assert(which in (2, 4),
                        "cannot use '?' in the %r field", _attribute[which])
            return None, _end

        # last day of the month
        if entry == 'l':
            _assert(which == 2,
                    "you can only specify a bare 'L' in the 'day' field")
            return None, _end

        # for the last 'friday' of the month, for example
        elif entry.startswith('l'):
            _assert(which == 4,
                    "you can only specify a leading 'L' in the 'weekday' field")
            es, _, ee = entry[1:].partition('-')
            _assert((entry[1:].isdigit() and 0 <= int(es) <= 7) or
                    (_ and es.isdigit() and ee.isdigit() and 0 <= int(es) <= 7 and 0 <= int(ee) <= 7),
                    "last <day> specifier must include a day number or range in the 'weekday' field, you entered %r", entry)
            return None, _end

        increment = None
        # increments
        if '/' in entry:
            entry, increment = entry.split('/')
            increment = int(increment, 10)
            _assert(increment > 0,
                    "you can only use positive increment values, you provided %r",
                    increment)

        # allow Sunday to be specified as weekday 7
        if which == 4:
            _end_limit = 7

        # handle singles and ranges
        good = _parse_piece(entry)

        # change Sunday to weekday 0
        if which == 4 and 7 in good:
            good.discard(7)
            good.add(0)

        return good, _end


class CrontabTrigger(object):
    __slots__ = 'matchers',

    def __init__(self, crontab):
        # check if the job_trigger is illegal
        configs = re.split("\s+", crontab)
        for config in configs:
            if config != "*" and not config.isdigit():
                raise ValueError("crontab expression error: %s" % crontab)
        self.matchers = self._make_matchers(crontab)

    def _make_matchers(self, crontab):
        """
        This constructs the full matcher struct.
        """
        crontab = _aliases.get(crontab, crontab)
        matchers = [_Matcher(which, entry)
                    for which, entry in enumerate(crontab.split())]

        if len(matchers) == 5:
            matchers.append(_Matcher(5, '*'))
        _assert(len(matchers) == 6,
                "improper number of cron entries specified")

        return Matcher(*matchers)

    def _test_match(self, index, dt):
        """
        This tests the given field for whether it matches with the current
        datetime object passed.
        """
        at = _attribute[index]
        attr = getattr(dt, at)
        if at == 'isoweekday':
            attr = attr() % 7
        return self.matchers[index](attr, dt)

    def get_next_runtime(self, start_time, equals=False):
        """
        获取下次需要触发的时间(返回时间戳为>=start_time 或者>start_time)
        :param start_time上次运行时间, 0表示无上次运行时间
        :param equals True表示返回结果>=start_time, False表示返回结果>start_time
        :return: timestamp，0表示不再执行
        """
        if equals:
            start_time = start_time - 1
        start_time = datetime.fromtimestamp(start_time)
        # handle timezones if the datetime object has a timezone and get a
        # reasonable future/past start time
        future = start_time.replace(second=0, microsecond=0) + _increments[0]()
        if future < start_time:
            # we are going backwards...
            _test = lambda: future.year < self.matchers.year
            if start_time.second or start_time.microsecond:
                future = start_time.replace(second=0, microsecond=0)
        else:
            # we are going forwards
            _test = lambda: self.matchers.year < future.year

        # Start from the year and work our way down. Any time we increment a
        # higher-magnitude value, we reset all lower-magnitude values. This
        # gets us performance without sacrificing correctness. Still more
        # complicated than a brute-force approach, but also orders of
        # magnitude faster in basically all cases.
        to_test = 5
        while to_test >= 0:
            if not self._test_match(to_test, future):
                inc = _increments[to_test](future, self.matchers)
                future += inc
                for i in xrange(0, to_test):
                    future = _increments[6 + i](future, inc)
                if _test():
                    return 0
                to_test = 5
                continue
            to_test -= 1

        # verify the match
        match = [self._test_match(i, future) for i in xrange(6)]
        if not all(match):
            sys.stderr.write("there is a bug in Crontab, please replace another one")
            return 0
        return int(time.mktime(future.timetuple()))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-27
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master.scheduler_rule.base import AbstractRule
import random


class RandomRule(AbstractRule):
    """
    Random rule of choosing worker
    """

    def choose(self, workers):
        if workers:
            return random.choice(workers)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-22
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from sqlalchemy import *
from sqlalchemy.exc import OperationalError
from angel_master.entity.worker import Worker
from traceback import format_exc
from threading import Thread
import time
import hashlib


def except_logger(func):
    """
    Decorator for logging when raising exception
    when a exception raised, the decorator will log it and raise again
    """
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except OperationalError as e:
            self.logger.error("couldn't execute db engine, %s" % str(e))
            # TODO report an incident to the admin
            raise
        except Exception as e:
            self.logger.warn(format_exc())
            raise
    return wrapper

worker_t = None
wg_t = None


class ResourceManager(Thread):
    """
    Implementation of resource manager
    """

    def __init__(self, logger, loop_time, timeout, engine, name, pwd):
        super(ResourceManager, self).__init__()
        global worker_t
        global wg_t
        self.started = False
        self.logger = logger
        self.loop_time = loop_time
        self.on_worker_remove = None
        self.timeout = timeout
        # obtain the hash of name
        self.name = hashlib.sha1(name.encode("utf8")).hexdigest()
        # obtain the hash of password
        self.pwd = hashlib.sha1(pwd.encode("utf8")).hexdigest()
        self.engine = engine
        metadata = MetaData(engine)
        worker_t = Table("worker", metadata,
                        Column('id', Integer, primary_key = True),
                        Column('group_id', Integer),
                        Column('login_time', Integer),
                        Column('desc', String(128)))
        wg_t = Table("worker_group", metadata,
                     Column("id", Integer, primary_key=True),
                     Column("name", String(64), nullable=False),
                     Column("desc", String(128)))
        metadata.create_all()
        # key is group id and value is list of the worker
        self.worker_groups = {}
        # key is worker_id and value is worker
        self.workers = {}

    def stop(self):
        self.started = False

    def run(self):
        """
        Main loop for scheculer
        """
        self.started = True
        while self.started:
            # sleep for a while
            time.sleep(self.loop_time)
            # calculate the timeout thredshold time
            thredshold = int(time.time()) - self.timeout
            # filter all workers using the thredshold time
            workers = list(filter(lambda x:x.refresh_time < thredshold, self.workers.values()))
            for worker in workers:
                self.logger.debug("%s timeout" % worker)
                # remove the worker
                del self.workers[worker.id]
                if worker.group_id in self.worker_groups:
                    ll = self.worker_groups[worker.group_id]
                    if worker in ll:
                        ll.remove(worker)
                # send message to listener
                if self.on_worker_remove:
                    self.on_worker_remove(worker)

    def heartbeat(self, worker_id, params):
        # find the worker
        worker = self.workers.get(worker_id)
        if not worker:
            self.login(self.name, self.pwd, worker_id)
        # update the parameters
        worker.refresh_time = int(time.time())
        for k,v in params.items():
            if k in worker.__slots__:
                setattr(worker, k, v)

    @except_logger
    def register(self, name, pwd, params):
        # check if the name and pwd is legal
        if self.name == name and self.pwd == pwd:
            # insert into database
            now = int(time.time())
            params["login_time"] = now
            res = worker_t.insert().execute(params)
            ids = res.inserted_primary_key
            if len(ids) == 1:
                return ids[0]
            raise Exception("register fail caused by %d records were insert" % len(ids))
        raise Exception("register fail caused by password or name is wrong")

    @except_logger
    def login(self, name, pwd, worker_id):
        if self.name == name and self.pwd == pwd:
            # check if the worker is register
            executable = select([worker_t.c.group_id]).where(worker_t.c.id == worker_id)
            res = self.engine.execute(executable)
            result = res.fetchall()
            if result and result[0]:
                # set the worker into data structure
                gid = result[0][0]
                now = int(time.time())
                worker = Worker(worker_id=worker_id, group_id=gid, refresh_time=now)
                self.workers[worker_id] = worker
                if gid not in self.worker_groups:
                    self.worker_groups[gid] = []
                ll = self.worker_groups[gid]
                if worker not in ll:
                    ll.append(worker)
                # update the worker's last login time
                worker_t.update().where(worker_t.c.id==worker_id).values(login_time=now)
                # info for debug
                self.logger.info("worker(wid=%d) login successfully" % worker_id)
            else:
                raise Exception("login fail caused by not register before")
        else:
            raise Exception("login fail caused by password or name is wrong")

    def logout(self, worker_id):
        # remove the worker out of data structure
        if worker_id in self.workers:
            worker = self.workers[worker_id]
            del self.workers[worker_id]
            if worker.group_id in self.worker_groups:
                ll = self.worker_groups[worker.group_id]
                if worker in ll:
                    ll.remove(worker)
                    if not ll:
                        del self.worker_groups[worker.group_id]
        # send a message to the listener
        if self.worker_listener:
            self.worker_listener.on_worker_remove(worker)

    def get_workers(self, group_id=None):
        """
        Obtain alive workers with the given group_id
        Return all alive workers when group_id is None
        """
        if group_id:
            return self.worker_groups.get(group_id)
        else:
            return list(self.workers.values())

    def get_worker(self, worker_id):
        """
        Return the alive worker with the given worker_id
        """
        return self.workers.get(worker_id)



from collections import deque
from collections import OrderedDict


class DAG(object):
    """ Directed acyclic graph implementation. """

    def __init__(self):
        """ Construct a new DAG with no nodes or edges. """
        self.reset_graph()

    def add_node(self, node_name):
        """ Add a node if it does not exist yet, or error out. """
        if node_name in self.graph:
            raise KeyError('node %s already exists' % node_name)
        self.graph[node_name] = set()
        self.back_graph[node_name] = set()

    def add_node_if_not_exists(self, node_name):
        if node_name not in self.graph:
            self.add_node(node_name)

    def delete_node(self, node_name):
        """ Deletes this node and all edges referencing it. """
        if node_name not in self.graph:
            raise KeyError('node %s does not exist' % node_name)
        children = self.graph.pop(node_name)
        parents = self.back_graph.pop(node_name)
        for child in children:
            self.back_graph[child].remove(node_name)
        for parent in parents:
            self.graph[parent].remove(node_name)

    def delete_node_if_exists(self, node_name):
        if node_name in self.graph:
            self.delete_node(node_name)

    def add_edge(self, ind_node, dep_node):
        """ Add an edge (dependency) between the specified nodes. """
        if ind_node not in self.graph or dep_node not in self.graph:
            raise KeyError('one or more nodes do not exist in graph')
        self.graph[ind_node].add(dep_node)
        self.back_graph[dep_node].add(ind_node)

    def delete_edge(self, ind_node, dep_node):
        """ Delete an edge from the graph. """
        if dep_node not in self.graph.get(ind_node, []):
            raise KeyError('this edge does not exist in graph')
        self.graph[ind_node].remove(dep_node)
        self.back_graph[dep_node].remove(ind_node)

    def predecessors(self, node):
        """ Returns a list of all predecessors of the given node """
        return self.back_graph.get(node)

    def downstream(self, node):
        """ Returns a list of all nodes this node has edges towards. """
        return self.graph.get(node)

    def all_downstreams(self, node):
        """Returns a set of all nodes ultimately downstream
        of the given node in the dependency graph, in
        topological order."""
        nodes = [node]
        nodes_seen = set()
        i = 0
        while i < len(nodes):
            downstreams = self.downstream(nodes[i])
            for downstream_node in downstreams:
                if downstream_node not in nodes_seen:
                    nodes_seen.add(downstream_node)
                    nodes.append(downstream_node)
            i += 1
        return nodes_seen

    def all_nodes(self):
        """ Return a list of all nodes """
        return self.graph.keys()

    def contains_node(self, node):
        """ Return whether the given node exist in the graph"""
        return node in self.graph

    def all_edges(self):
        """ Return a list of all edges """
        result = set()
        for parent, children in self.graph.items():
            for child in children:
                result.add((parent, child))
        return result

    def all_leaves(self):
        """ Return a list of all leaves (nodes with no downstreams) """
        return set(key for key in self.graph if not self.graph[key])

    def ind_nodes(self):
        """ Returns a list of all nodes in the graph with no dependencies. """
        return set(key for key in self.back_graph if not self.back_graph[key])

    def validate(self):
        """ Returns (Boolean, message) of whether DAG is valid. """
        nodes = self.topological_sort()
        return len(nodes) == len(self.graph)

    def topological_sort(self):
        """ Returns a topological ordering of the DAG.
        Raises an error if this is not possible (graph is not valid).
        """
        in_degree = {}
        for u in self.graph:
            in_degree[u] = 0
        for u in self.graph:
            for v in self.graph[u]:
                in_degree[v] += 1
        queue = deque()
        for u in in_degree:
            if in_degree[u] == 0:
                queue.appendleft(u)
        l = []
        while queue:
            u = queue.pop()
            l.append(u)
            for v in self.graph[u]:
                in_degree[v] -= 1
                if in_degree[v] == 0:
                    queue.appendleft(v)
        return l

    def clone(self):
        """Returns a dag with the same node instance"""
        result = DAG()
        for k, v in self.graph.items():
            result.graph[k] = v.copy()
        for k, v in self.back_graph.items():
            result.back_graph[k] = v.copy()
        return result

    def size(self):
        return len(self.graph)

    def reset_graph(self):
        """ Restore the graph to an empty state. """
        self.graph = dict()
        self.back_graph = dict()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-29
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master.api.base_service import AbstractService
from traceback import format_exc

CODE_SUCCESS = 0
CODE_FAIL = 1
on_logs = dict()
off_workers = set()


def exception_hander(func):
    """
    Decorator for obtaining the response after calling service
    When a exception raised, the response returns CODE_ERROR
    """
    def wrapper(self, *args, **kwargs):
        resp = {}
        try:
            result = func(self, *args, **kwargs)
            resp["code"] = CODE_SUCCESS
            if result is not None:
                resp["content"] = result
        except Exception as e:
            resp["code"] = CODE_FAIL
            resp["content"] = str(e)
            self.logger.warn(format_exc())
        return resp
    return wrapper


class RPCService(AbstractService):
    """
    Implementation of api service
    """

    def __init__(self, scheduler, resourcemanager, logger):
        self.scheduler = scheduler
        self.resourcemanager = resourcemanager
        self.logger = logger

    @exception_hander
    def login(self, name, password, worker_id):
        return self.resourcemanager.login(name, password, worker_id)

    @exception_hander
    def register(self, name, password, params):
        return self.resourcemanager.register(name, password, params)

    @exception_hander
    def logout(self, worker_id):
        return self.resourcemanager.logout(worker_id)

    @exception_hander
    def all_alive_workers(self):
        workers = self.resourcemanager.get_workers()
        result = []
        for worker in workers:
            result.append({"id":worker.id, "cpu_free":worker.cpu_free,
                           "memory_free":worker.memory_free, "disk_write":worker.disk_write,
                           "disk_read":worker.disk_read, "net_send":worker.net_send,
                           "net_rev":worker.net_rev, "running_tasks":worker.running_tasks,
                           "refresh_time":worker.refresh_time})
        return result

    @exception_hander
    def heartbeat(self, worker_id, params):
        return self.resourcemanager.heartbeat(worker_id, params)

    @exception_hander
    def pull_tasks(self, worker_id):
        return self.scheduler.pull_tasks(worker_id)

    @exception_hander
    def pull(self, worker_id):
        if worker_id in off_workers:
            off_workers.remove(worker_id)
            return {"cmd":"shutdown"}
        result = {}
        tasks = self.scheduler.pull_tasks(worker_id)
        if tasks:
            result["tasks"] = tasks
        if worker_id in on_logs:
            result["logs"] = list(on_logs.pop(worker_id))
        return result

    @exception_hander
    def task_callback(self, task_id, done_time, worker_id, state):
        return self.scheduler.task_callback(task_id, done_time, worker_id, state)

    @exception_hander
    def add_log(self, task_id):
        task_id = int(task_id)
        task = self.scheduler.get_task(task_id)
        if task is None or task.worker_id is None:
            raise Exception("the task(id=%d) is not running" % task_id)
        worker_id = task.worker_id
        if self.resourcemanager.get_worker(worker_id) is None:
            raise Exception("The worker running the task is death")
        if worker_id in on_logs:
            on_logs[worker_id].add(task_id)
        else:
            on_logs[worker_id] = set([task_id])

    @exception_hander
    def apply_job_group(self, group_id):
        self.scheduler.delete_job_group(group_id)
        self.scheduler.add_job_group(group_id)

    @exception_hander
    def delete_job_group(self, group_id):
        self.scheduler.delete_job_group(group_id)

    @exception_hander
    def get_pending_tasks(self):
        return self.scheduler.pending_tasks()

    @exception_hander
    def get_alive_tasks(self):
        return self.scheduler.alive_tasks()

    @exception_hander
    def get_alive_job_groups(self):
        return self.scheduler.alive_job_groups()

    @exception_hander
    def get_alive_task_groups(self):
        return self.scheduler.alive_task_groups()

    @exception_hander
    def shutdown_workers(self, workers):
        off_workers = off_workers.union(set(workers))




# @Time         2017-11-29
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93



class AbstractService(object):
    """
    Implementation of api service
    """

    def __init__(self, scheduler, resourcemanager, logger):
        self.scheduler = scheduler
        self.resourcemanager = resourcemanager
        self.logger = logger

    def login(self, name, password, worker_id):
        pass

    def register(self, name, password, params):
        pass

    def logout(self, worker_id):
        pass

    def all_alive_workers(self):
        pass

    def heartbeat(self, worker_id, params):
        pass

    def pull_tasks(self, worker_id):
        pass

    def pull(self, worker_id):
        pass

    def task_callback(self, task_id, done_time, worker_id, state):
        pass

    def add_log(self, task_id):
        pass

    def apply_job_group(self, group_id):
        pass

    def delete_job_group(self, group_id):
        pass

    def get_pending_tasks(self):
        pass

    def get_alive_tasks(self):
        pass

    def get_alive_job_groups(self):
        pass

    def get_alive_task_groups(self):
        pass

    def shutdown_workers(self, workers):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master.entity.task import Task
from angel_master.core.dag import DAG
from angel_master.trigger.crontab import CrontabTrigger
from angel_master.entity.task_group import TaskGroup
from threading import Thread, Lock
import time


class DAGScheduler(object):
    def __init__(self, logger, jobstore):
        self.logger = logger
        self.jobstore = jobstore
        self.task_lock = Lock()
        self.job_lock = Lock()
        self.groups = {}
        self.job_dag = DAG()
        self.task_dag = DAG()
        self.job_groups = {}
        self.task_groups = {}
        self.jobs = {}
        self.tasks = {}

    def initialize(self, groups, jobs, job_dependences):
        if not groups:
            return
        # validate data
        _jobs = {}
        _job_groups = {}
        _job_depens = {}
        _illegal_groups = set()
        for group in groups:
            _job_groups[group.id] = []
            _job_depens[group.id] = []
        for job in jobs:
            _job_groups[job.group_id].append(job)
            _jobs[job.id] = job
            try:
                job.trigger = CrontabTrigger(job.trigger)
            except Exception as e:
                # TODO send a report
                self.logger.error("the trigger of  %s is illegal" % str(job))
                _illegal_groups.add(job.group_id)
        for group in groups:
            if not group.first_job_id or _jobs[group.first_job_id].group_id != group.id:
                # TODO send a report
                self.logger.error("the first job of the group %s is illegal" % str(group))
                _illegal_groups.add(group.id)
        for job_depend in job_dependences:
            job1 = _jobs[job_depend.job1]
            job2 = _jobs[job_depend.job2]
            if job_depend.group_id == job1.group_id == job2.group_id:
                _job_depens[job_depend.group_id].append((job1, job2))
            else:
                # TODO send a report
                self.logger.error("the job_depend %s is out of group" % str(job_depend))
                _illegal_groups.add(job_depend.group_id)
        # add job groups into jobs, job_groups, job_dag
        for group in groups:
            if group.id not in _illegal_groups:
                self._add_job_group(group, _job_groups[group.id], _job_depens[group.id])
        # release the memory
        del _jobs
        del _job_depens
        del _job_groups
        del _illegal_groups
        # remove cycles from job dag
        self._remove_job_cycles()
        # build task dag
        for group in groups:
            if group.id in self.job_groups:
                self._build_task_group(group.id)

    def _add_job_group(self, group, jobs, job_dependences):
        """
        Helper method for add a job group
        """
        with self.job_lock:
            if group.id in self.groups:
                # the group is already exist
                return
            self.groups[group.id] = group
            self.job_groups[group.id] = jobs
            for job in jobs:
                self.jobs[job.id] = job
                self.job_dag.add_node(job)
            for job_depend in job_dependences:
                self.job_dag.add_edge(job_depend[0], job_depend[1])

    def _remove_job_group(self, group_id):
        """
        Help method for remove a job group
        """
        with self.job_lock:
            if group_id not in self.groups:
                return
            del self.groups[group_id]
            jobs = self.job_groups.pop(group_id)
            for job in jobs:
                del self.jobs[job.id]
                self.job_dag.delete_node_if_exists(job)

    def _remove_job_cycles(self):
        """
        Helper method to remove cycle from job dag
        """
        with self.job_lock:
            # check if the job dag is acyclic
            if not self.job_dag.validate():
                topological_sort = self.job_dag.topological_sort()
                cycle_jobs = filter(lambda x: x not in topological_sort,
                                    self.jobs.values())
                group_ids = set(map(lambda x: x.group_id, cycle_jobs))
                for group_id in group_ids:
                    # TODO send a report
                    self.logger.error("jobs in the group(id=%d) is acyclic" % group_id)
                    sub_jobs = self.job_groups[group_id]
                    del self.job_groups[group_id]
                    del self.groups[group_id]
                    for job in sub_jobs:
                        del self.jobs[job.id]
                        self.job_dag.delete_node_if_exists(job)

    def _build_task_group(self, group_id):
        """
        Helper method to build new task by the given group_id
        """
        with self.job_lock:
            if group_id not in self.groups:
                return
            job_group = self.groups[group_id]
            now = int(time.time())
            jobs = self.job_groups[group_id]
            first_job = self.jobs[job_group.first_job_id]
            start_time = first_job.trigger.get_next_runtime(now, True)
            with self.task_lock:
                # check if the task group is already exist
                for item in self.task_groups.keys():
                    if item.runtime == start_time and item.group_id == group_id:
                        # TODO the task group is already exist
                        return
                task_group = TaskGroup(group_id=group_id, runtime=start_time, state=Task.STATE_INIT)
                self.jobstore.insert_task_group(task_group)
                self.task_groups[task_group] = []
                job_task_map = {}
                # new task from job and set into tasks,task_groups,task_dag
                for job in jobs:
                    runtime = job.trigger.get_next_runtime(start_time, True)
                    task = Task(job_id=job.id, group_id=task_group.id, runtime=runtime,
                                running_timeout=job.running_timeout)
                    task.group = task_group
                    # insert the task
                    self.jobstore.insert_task(task)
                    # add task node into task_dag, tasks, task_groups
                    self.tasks[task.id] = task
                    self.task_groups[task_group].append(task)
                    job_task_map[job] = task
                    self.task_dag.add_node(task)
                    # TODO dubug
                    msg = str(task) + " will be executed at " + time.strftime("%m-%d %H:%M",
                                                                              time.localtime(task.runtime))
                    self.logger.debug(msg)
                # add task edge into task_dag
                for job in self.job_groups[group_id]:
                    task1 = job_task_map[job]
                    for job2 in self.job_dag.downstream(job):
                        task2 = job_task_map.get(job2)
                        if task2:
                            self.task_dag.add_edge(task1, task2)

    def _remove_task_group(self, job_group_id):
        """
        Helper method to remove task group by given job_group_id
        """
        with self.task_lock:
            groups = list(filter(lambda x: x.group_id == job_group_id,
                          self.task_groups.keys()))
            for group in groups:
                tasks = self.task_groups.pop(group)
                for task in tasks:
                    del self.tasks[task.id]
                    self.task_dag.delete_node_if_exists(task)

    def get_pending_tasks(self):
        """
        Return independent task, such as task in STATE_INIT, STATE_RUNNING
        """
        tasks = self.task_dag.ind_nodes()
        tmp = filter(lambda t: t.state == Task.STATE_INIT or t.state == Task.STATE_TIMING or
                     t.state == Task.STATE_RUNNING, tasks)
        return list(tmp)

    def is_task_group_finish(self, group):
        """
        Return if the tasks in the group are all success
        """
        tasks = self.task_groups.get(group)
        if tasks:
            for task in tasks:
                if task.state != Task.STATE_SUCCESS:
                    return False
            return True
        return False

    def on_task_success(self, task):
        """
        Remove the task after the task is success
        """
        del self.tasks[task.id]
        self.task_dag.delete_node_if_exists(task)
        if self.is_task_group_finish(task.group):
            # record the task group
            self.jobstore.update_task_group(task.group.id, state=Task.STATE_SUCCESS)
            # rebuild new task group
            self._build_task_group(task.group.group_id)
            del self.task_groups[task.group]

    def on_task_not_success(self, task):
        """
        Record the state of the task and alert
        """
        # revord the task group
        self.jobstore.update_task_group(task.group_id, state=Task.STATE_FAIL)
        # rebuild new task group
        self._build_task_group(task.group.group_id)

    def get_task(self, task_id):
        """
        Obtain task object by given task_id
        """
        return self.tasks.get(task_id)

    def add_job_group(self, group_id):
        """
        Force to add a job group by the given group id
        Besides, insert a new task dag
        """
        # encapsulate data and validate
        group, jobs, job_dependences = self.jobstore.query_all(group_id)
        _jobs = {}
        for job in jobs:
            try:
                job.trigger = CrontabTrigger(job.trigger)
            except Exception as e:
                # TODO the trigger of the job is illegal
                raise Exception("the trigger of  %s is illegal" % str(job))
            _jobs[job.id] = job
        if group.first_job_id not in _jobs:
            # TODO the first job id of the group is illegal
            raise Exception("the first job of the group %s is illegal" % str(group))
        _job_depends = []
        for job_depend in job_dependences:
            job1 = _jobs.get(job_depend.job1)
            job2 = _jobs.get(job_depend.job2)
            if job1 and job2 and job1.group_id == job2.group_id:
                _job_depends.append((job1, job2))
            else:
                # TODO the job dependence is illegal
                raise Exception('the job depend %s isillegal' % str(job_depend))
        # add into job dag
        self._add_job_group(group, jobs, _job_depends)
        # remove cycle from job dag if exist
        self._remove_job_cycles()
        # build a new task dag
        self._build_task_group(group.id)

    def delete_job_group(self, group_id):
        """
        Force to remove a job group by the given group id
        Besides, remove the tasks related
        """
        self._remove_job_group(group_id)
        self._remove_task_group(group_id)


class Scheduler(Thread):
    """
    A Scheduler that run in loops for time scheduler
    """

    def __init__(self, jobstore, resource_manager, scheduler_rule, logger,
                 scheduler_timeout, loop_time):
        super(Scheduler, self).__init__()
        self.jobstore = jobstore
        self.resource_manager = resource_manager
        self.scheduler_rule = scheduler_rule
        self.logger = logger
        self.scheduler_timeout = scheduler_timeout
        self.loop_time = loop_time
        self.started = False
        job_groups, jobs, job_dependences = self.jobstore.query_all()
        self.timing_tasks = []
        self.running_timeout_tasks = []
        self.dag_scheduler = DAGScheduler(logger, jobstore)
        self.dag_scheduler.initialize(job_groups, jobs, job_dependences)
        self.resource_manager.on_worker_remove = self.on_worker_remove

    def stop(self):
        self.started = False

    def run(self):
        """
        Main loop for time scheduler
        """
        self.started = True
        # check the status
        while self.started:
            pending_tasks = self.dag_scheduler.get_pending_tasks()
            # classify tasks according to their runtime
            running_timeout_tasks = []
            schedule_timeout_tasks = []
            timing_tasks = []
            now = int(time.time())
            schedule_timeout_time = now - self.scheduler_timeout
            for task in pending_tasks:
                if task.state == Task.STATE_INIT or task.state == Task.STATE_TIMING:
                    if task.runtime < schedule_timeout_time:
                        schedule_timeout_tasks.append(task)
                    elif task.runtime <= now:
                        timing_tasks.append(task)
                elif task.state == Task.STATE_RUNNING:
                    if task.runtime <= now - task.running_timeout:
                        running_timeout_tasks.append(task)
            # handle running timeout tasks
            self._handle_running_timeout(running_timeout_tasks)
            # handle schedule timeout tasks
            self._handle_schedule_timeout(schedule_timeout_tasks)
            # handle timing tasks
            self._handle_timing_tasks(timing_tasks)
            # sleep a moment
            time.sleep(self.loop_time)

    def _handle_running_timeout(self, tasks):
        """
        Handle running timeout tasks
        """
        now = int(time.time())
        for task in tasks:
            task.state = Task.STATE_RUNNING_TIMEOUT
            task.done_time = now
            # record the task state
            self.jobstore.update_task(task.id, state=Task.STATE_RUNNING_TIMEOUT, done_time=now)
            self.dag_scheduler.on_task_not_success(task)
            # info for debug
            self.logger.info("%s is running timeout" % str(task))
        self.running_timeout_tasks = tasks

    def _handle_schedule_timeout(self, tasks):
        """
        Handle schedule timeout tasks
        """
        now = int(time.time())
        for task in tasks:
            # reset its worker
            task.worker_id = None
            task.state = Task.STATE_SCHEDULE_TIMEOUT
            task.done_time = now
            # record the task state
            self.jobstore.update_task(task.id, state=Task.STATE_SCHEDULE_TIMEOUT, done_time=now)
            self.dag_scheduler.on_task_not_success(task)
            # TODO debug
            self.logger.debug("%s schedule timeout" % str(task))

    def _handle_timing_tasks(self, tasks):
        """
        Handle timing tasks
        """
        for task in tasks:
            # check if it has worker
            if task.worker_id:
                continue
            # get related worker list from resourceManager
            worker_gid = self.dag_scheduler.jobs[task.job_id].worker_gid
            worker_list = self.resource_manager.get_workers(worker_gid)
            # choose a worker according to scheduler rule
            worker = self.scheduler_rule.choose(worker_list)
            if worker:
                task.worker_id = worker.id
            # TODO debug
            self.logger.debug("%s is timming" % str(task))
        self.timing_tasks = tasks

    def get_task(self, task_id):
        """
        Return the task with the given task_id
        """
        return self.dag_scheduler.tasks.get(task_id)

    def pull_tasks(self, worker_id):
        """
        Return the tasks that should be executed and killed
        """
        # obtain the tasks that should be executed
        timing_tasks = list(filter(lambda task: task.worker_id == worker_id, self.timing_tasks))
        # obtain the tasks that should be terminated
        timeout_tasks = list(filter(lambda task: task.worker_id == worker_id, self.running_timeout_tasks))
        result = []
        for task in timing_tasks:
            # record the state STATE_RUNNING
            self.jobstore.update_task(task.id, state=Task.STATE_RUNNING, worker_id=worker_id)
            task.state = Task.STATE_RUNNING
            # serialize task in dict
            result.append({"id":task.id, "job_id":task.job_id, "state":Task.STATE_RUNNING})
        for task in timeout_tasks:
            # serialize task in dict
            result.append({"id":task.id, "job_id":task.job_id, "state":Task.STATE_RUNNING_TIMEOUT})
        return result

    def task_callback(self, task_id, done_time, worker_id, state):
        """
        Callback from worker when the task has been finish(success/fail)
        """
        # TODO for debug
        self.logger.debug("Task(id=%s,wid=%s,state=%s) callback" % (task_id, worker_id, state))
        # record the task state
        self.jobstore.update_task(task_id, state=state, done_time=done_time)
        task = self.dag_scheduler.get_task(task_id)
        if task:
            task.state = state
            # dag schedule
            if state == Task.STATE_SUCCESS:
                self.dag_scheduler.on_task_success(task)
            else:
                self.dag_scheduler.on_task_not_success(task)

    def add_job_group(self, group_id):
        """
        Add a job group from DB by the given group_id
        """
        self.dag_scheduler.add_job_group(group_id)

    def delete_job_group(self, group_id):
        """
        Delete a job group by the given group_id
        """
        self.dag_scheduler.delete_job_group(group_id)

    def pending_tasks(self):
        tasks = self.dag_scheduler.get_pending_tasks()
        result = []
        for task in tasks:
            result.append(task.to_dict())
        return result

    def alive_tasks(self):
        tasks = self.dag_scheduler.tasks.values()
        result = []
        for task in tasks:
            result.append(task.to_dict())
        return result

    def alive_job_groups(self):
        groups = self.dag_scheduler.groups.values()
        result = []
        for group in groups:
            result.append(group.to_dict())
        return result

    def alive_task_groups(self):
        groups = self.dag_scheduler.task_groups.keys()
        result = []
        for group in groups:
            result.append(group.to_dict())
        return result

    def on_worker_remove(self, worker):
        """
        Callback from resourceManager when the worker is timeout
        """
        tasks = filter(lambda task: task.worker_id == worker.id,
                       self.timing_tasks)
        for task in tasks:
            task.worker_id = None



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-29
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master.api.base_server import AbstractServer
from angel_master.api.rpc_service import RPCService
from xmlrpc.server import SimpleXMLRPCServer
from socketserver import ThreadingMixIn


class ThreadingXMLRPCServer(ThreadingMixIn, SimpleXMLRPCServer):
    pass


class XMLRPCServer(AbstractServer):
    """
    RPCServer implemented with xmlrpc.server
    """

    def __init__(self, host, port, logger, scheduler, resourcemanager):
        super(XMLRPCServer, self).__init__(host, port, logger)
        rpc_service = RPCService(scheduler, resourcemanager, logger)
        #self.server = SimpleXMLRPCServer((host, int(port)), logRequests=False)
        self.server = ThreadingXMLRPCServer((host, port), logRequests=False, allow_none=True)
        self.server.register_instance(rpc_service)

    def start(self):
        """
        Startup the service
        """
        try:
            self.server.serve_forever()
        except Exception as e:
            self.logger.warn("rpc server start fail caused by %s" % str(e))
            raise

    def stop(self):
        """
        Stop the service
        """
        try:
            self.server.server_close()
        except Exception as e:
            self.logger.warn("rpc server stop fail caused by %s" % str(e))
            raise



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-27
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


class AbstractRule(object):
    """
    Interface of Scheduler rule
    """

    def choose(self, workers):
        """
        Choose one of the given workers, to execute task
        """
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-27
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master.resource_manager import ResourceManager
from angel_master.scheduler import Scheduler
from angel_master.jobstore import JobStore
from angel_master.scheduler_rule.rand import RandomRule
from angel_master.core import config_util
from angel_master.api.rpc_server import XMLRPCServer
from sqlalchemy import create_engine
from traceback import format_exc
import logging
import logging.config
import os

ETC_PATH = "/etc/angel/master/"

scheduler = None
resource_manager = None
jobstore = None
logger = None
rpc_server = None


def serve_forever(debug=False):
    global scheduler
    global resource_manager
    global jobstore
    global logger
    global rpc_server
    # setup logger
    logger_config_path = os.path.join(ETC_PATH, "logger.conf")
    logging.config.fileConfig(logger_config_path)
    if debug:
        logger = logging.getLogger("debug")
    else:
        logger = logging.getLogger("release")
    # load setting
    config_path = os.path.join(ETC_PATH, "master.conf")
    if not os.path.isfile(config_path):
        logger.error("configuration file %s not exist" % config_path)
    try:
        # import settings
        config = config_util.parse(config_path, 'master')
        name = config['name']
        pwd = config['password']
        host = config["host"]
        port = int(config["port"])
        db_url = config["db_url"]
        max_overflow = int(config["max_overflow"])
        heartbeat_loop_time = int(config["heartbeat_loop_time"])
        worker_timeout = int(config["worker_timeout"])
        schedule_timeout = int(config["schedule_timeout"])
        schedule_loop_time = int(config["schedule_loop_time"])
    except Exception as e:
        logger.error("config file '%s' error: %s" %(config_path, format_exc()))
        raise
    # init db engine
    engine = create_engine(db_url, max_overflow=max_overflow)
    # init jobstore
    jobstore = JobStore(engine, logger)
    # init resource manager
    resource_manager = ResourceManager(logger, heartbeat_loop_time, worker_timeout, engine, name, pwd)
    # init scheduler
    rule = RandomRule()
    scheduler = Scheduler(jobstore, resource_manager, rule, logger, schedule_timeout, schedule_loop_time)
    # init rpc service
    rpc_server = XMLRPCServer(host, port, logger, scheduler, resource_manager)
    # start services of Master
    _start()


def _start():
    # startup resource manager
    resource_manager.start()
    # start up scheduler
    scheduler.start()
    # start up rpc service
    rpc_server.start()


def _stop():
    # stop rpc service
    rpc_server.stop()
    # stop scheduler
    scheduler.stop()
    # stop resource manager
    resource_manager.stop()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

class Job(object):
    """
    Contains the properties of Job, maps with the table 'job' in DB
    """
#    __slots__ = ("id", "group_id", "name", "desc", "worker_gid", "trigger", "expiry_time")
    def __init__(self, id=None, group_id=None, name=None, desc=None, worker_gid=None, trigger=None,
                 running_timeout=None):
        self.id = id
        self.group_id = group_id
        self.name = name
        self.desc = desc
        self.worker_gid = worker_gid
        self.trigger = trigger
        self.running_timeout = running_timeout

    def __str__(self):
        return "Job(id = %d, name = %s)" % (self.id, self.name)

    def __repr__(self):
        return "Job(id = %d, name = %s)" % (self.id, self.name)

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        return isinstance(other, Job) and self.id == other.id





#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

class JobGroup(object):
    """
    Contains the properties of JobGroup, maps with the table 'job_group' in DB
    """
#    __slots__ = ("id", "name", "desc", "first_job_id")

    def __init__(self, id=None, name=None, desc=None, first_job_id=None):
        self.id = id
        self.name = name
        self.desc = desc
        self.first_job_id = first_job_id

    def __str__(self):
        return "JobGroup(id = %d, name = %s)" % (self.id, self.name)

    def __repr__(self):
        return "JobGroup(id = %d, name = %s)" % (self.id, self.name)

    def __eq__(self, other):
        return isinstance(other, JobGroup) and self.id == other.id

    def to_dict(self):
        return {"id":self.id, "name":self.name, "desc":self.desc, "first_job_id":self.first_job_id}




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

class WorkerGroup(object):
    __slots__ = ("id", "name", "desc")

    def __str__(self):
        return "WorkerGroup(id = %d, name = %s)" % (self.id, self.name)

    def __repr__(self):
        return "WorkerGroup(id = %d, name = %s)" % (self.id, self.name)

    def __eq__(self, other):
        return isinstance(other, WorkerGroup) and self.id == other.id




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-27
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

import configparser
import os


def parse(filepath, section):
    if not os.path.isfile(filepath):
        raise Exception("file '%s' not exist" % filepath)
    parser = configparser.ConfigParser()
    parser.read(filepath)
    lists = parser.items(section)
    config = {}
    for ll in lists:
        if len(ll) != 2:
            raise Exception('the configuration of %s is illegal : %s' % (filepath, str(ll)))
        config[ll[0]] = ll[1]
    return config



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

import time


class Worker(object):
    """
    Contains the properties of Worker
    """
    __slots__ = ("id", "group_id", "cpu_free", "memory_free", "disk_write", "disk_read",
                 "net_send", "net_rev",  "running_tasks", "refresh_time", "desc")

    def __init__(self, worker_id=0, group_id=None, desc=None, refresh_time=int(time.time())):
        self.id = worker_id
        self.group_id = group_id
        self.desc = desc
        self.refresh_time = refresh_time
        self.cpu_free = 0
        self.memory_free = 0
        self.disk_write = 0
        self.disk_read= 0
        self.net_send = 0
        self.net_rev = 0
        self.running_tasks = 0

    def __str__(self):
        return "Worker(id = %s)" % self.id

    def __repr__(self):
        return "Worker(id = %s)" % self.id

    def __eq__(self, other):
        return isinstance(other, Worker) and self.id == other.id




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-29
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


class AbstractServer(object):
    """
    Abstract RPCServer for remote procedure calls
    """

    def __init__(self, host, port, logger):
        self.host = host
        self.port = port
        self.logger = logger

    def start(self):
        """
        Startup the service
        """
        pass

    def stop(self):
        """
        Stop the service
        """
        pass




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-24
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master.entity import *
from angel_master.entity.job import Job
from angel_master.entity.task import Task
from angel_master.entity.job_group import JobGroup
from angel_master.entity.job_depend import JobDepend
from angel_master.entity.task_group import TaskGroup
from sqlalchemy import *
from sqlalchemy.orm import mapper, sessionmaker
from sqlalchemy.exc import OperationalError
from traceback import format_exc

SessionBuilder = None


def except_logger(func):
    """
    Decorator for logging when raising exception
    when a exception raised, the decorator will log it and raise again
    """
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except OperationalError as e:
            self.logger.error("couldn't execute db engine, %s" % str(e))
            raise
        except Exception as e:
            self.logger.warn(format_exc())
            raise
    return wrapper



class JobStore(object):
    """
    Implementation of Jobstore
    """

    def __init__(self, engine, logger):
        global SessionBuilder
        SessionBuilder = sessionmaker(bind=engine)
        self.engine = engine
        self.logger = logger
        m = MetaData(engine)
        self.j_t = Table("job", m,
                         Column('id', Integer, primary_key=True),
                         Column('group_id', Integer, ForeignKey('job_group.id')),
                         Column('name', String(64), nullable=False),
                         Column('desc', String(128)),
                         Column('worker_gid', Integer),
                         Column('trigger', String(32)),
                         Column('running_timeout', Integer, nullable=False))
        self.t_t = Table("task", m,
                         Column('id', Integer, primary_key=True),
                         Column('job_id', Integer, ForeignKey('job.id')),
                         Column('group_id', Integer, ForeignKey('task_group.id')),
                         Column('runtime', Integer, nullable=False),
                         Column('worker_id', Integer),
                         Column('done_time', Integer),
                         Column('state', Integer, default=Task.STATE_INIT))
        self.tg_t = Table("task_group", m,
                          Column('id', Integer, primary_key=True),
                          Column('group_id', Integer, ForeignKey('job_group.id')),
                          Column('runtime', Integer, nullable=False),
                          Column('done_time', Integer),
                          Column('state', Integer, nullable=False, default=Task.STATE_INIT))
        self.jg_t = Table("job_group", m,
                          Column('id', Integer, primary_key=True),
                          Column('name', String(64), nullable=False),
                          Column('desc', String(128)),
                          Column('first_job_id', Integer, ForeignKey('job.id')))
        self.jd_t = Table("job_depend", m,
                          Column('group_id', Integer, ForeignKey('job_group.id')),
                          Column('job1', Integer, ForeignKey('job.id'), primary_key=True),
                          Column('job2', Integer, ForeignKey('job.id'), primary_key=True))
        m.create_all()
        mapper(JobGroup, self.jg_t)
        mapper(Job, self.j_t)
        mapper(JobDepend, self.jd_t)
        mapper(TaskGroup, self.tg_t)
        mapper(Task, self.t_t)

    @except_logger
    def insert_task_group(self, group):
        session = SessionBuilder()
        session.add(group)
        session.commit()
        group.id = group.id
        session.close()

    @except_logger
    def insert_task(self, task):
        session = SessionBuilder()
        session.add(task)
        session.commit()
        task.id = task.id
        session.close()

    @except_logger
    def update_task(self, task_id, **kwargs):
        executable=self.t_t.update().where(self.t_t.c.id==task_id).values(**kwargs)
        self.engine.execute(executable)

    @except_logger
    def update_task_group(self, group_id, **kwargs):
        executable=self.tg_t.update().where(self.tg_t.c.id==group_id).values(**kwargs)
        res = self.engine.execute(executable)

    @except_logger
    def query_all(self, group_id=None):
        session = SessionBuilder()
        if group_id:
            # query group
            groups = session.query(JobGroup).filter(JobGroup.id==group_id).one()
            if not groups:
                return (None, None, None)
            # query jobs
            jobs = session.query(Job).filter(Job.group_id==group_id).all()
            # query job_dependences
            job_depends = session.query(JobDepend).filter(JobDepend.group_id==group_id).all()
        else:
            # query group
            groups = session.query(JobGroup).all()
            if not groups:
                return (None, None, None)
            # query jobs
            jobs = session.query(Job).all()
            # query job_dependences
            job_depends = session.query(JobDepend).all()
        session.close()
        return (groups, jobs, job_depends)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


class TaskGroup(object):
    """
    Contains the properties of task, which is the instance of Job
    """

    #__slots__ = ("id", "group_id", "runtime", "done_time", "state")

    def __init__(self, id=None, group_id=None, runtime=None, state=None):
        self.id = id
        self.group_id = group_id
        self.runtime = runtime
        self.state = state

    def __str__(self):
        return "TaskGroup(id = %d, group_id = %d, runtime = %d)" % (self.id, self.group_id,
                                                                    self.runtime)

    def __repr__(self):
        return "TaskGroup(id = %d, group_id = %d, runtime = %d)" % (self.id, self.group_id,
                                                                    self.runtime)

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        return isinstance(other, TaskGroup) and self.id == other.id

    def to_dict(self):
        return {"id":self.id, "group_id":self.group_id, "runtime":self.runtime,
                "done_time":self.done_time, "state":self.state}




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-24
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

class JobDepend(object):
    """
    Job Dependence
    (job1, job2) means that the job2 has to run after job1
    """
#    __slots__ = ('job1', 'job2')

    def __init__(self, group_id=None, job1=None, job2=None):
        self.group_id = group_id
        self.job1 = job1
        self.job2 = job2

    def __str__(self):
        return "JobDepend(group_id=%d, job1=%d, job2=%d)" % (self.group_id, self.job1, self.job2)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-27
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_master import master
#from sqlalchemy import *
#from sqlalchemy.orm import sessionmaker, mapper
#from master.entity.job import Job
#from master.entity.task import Task
#from master.entity.task_group import TaskGroup
#from master.entity.job_group import JobGroup
#from master.entity.job_depend import JobDepend
#from master.jobstore.jobstore import JobStore

if __name__ !=  '__main__':
    exit()

#engine = create_engine('mysql+pymysql://root:root@127.0.0.1:3306/angel')
#m = MetaData(engine)
## mapper
#m = MetaData(engine)
#j_t = Table("job", m, Column('id', Integer, primary_key=True),
#            Column('group_id', Integer, ForeignKey('job_group.id')),
#            Column('name', String(64), nullable=False),
#            Column('desc', String(128)),
#            Column('worker_gid', Integer),
#            Column('trigger', String(32)),
#            Column('running_timeout', Integer, nullable=False))
#t_t = Table("task", m, Column('id', Integer, primary_key=True),
#            Column('job_id', Integer, ForeignKey('job.id')),
#            Column('group_id', Integer, ForeignKey('task_group.id')),
#            Column('runtime', Integer, nullable=False),
#            Column('worker_id', Integer),
#            Column('done_time', Integer),
#            Column('state', Integer, default=Task.INIT_STATE))
#tg_t = Table("task_group", m, Column('id', Integer, primary_key=True),
#             Column('group_id', Integer, ForeignKey('job_group.id')),
#             Column('runtime', Integer, nullable=False),
#             Column('done_time', Integer),
#             Column('state', Integer, nullable=False, default=Task.INIT_STATE))
#jg_t = Table("job_group", m, Column('id', Integer, primary_key=True),
#             Column('name', String(64), nullable=False),
#             Column('desc', String(128)),
#             Column('first_job_id', Integer))
#jd_t = Table("job_depend", m, Column('group_id', Integer, ForeignKey('job_group.id')),
#             Column('job1', Integer, ForeignKey('job.id'), primary_key=True),
#             Column('job2', Integer, ForeignKey('job.id'), primary_key=True))
#m.create_all()
#mapper(JobGroup, jg_t)
#mapper(Job, j_t)
#mapper(JobDepend, jd_t)
#mapper(TaskGroup, tg_t)
#mapper(Task, t_t)
#
#SessionBuilder = sessionmaker(engine)
#session = SessionBuilder()
## clean data
#session.query(Job).delete()
#session.query(JobGroup).delete()
#session.query(JobDepend).delete()
#session.commit()
#
## insert data
#jg1 = JobGroup(id=1, name='group1', first_job_id=1, desc='It is group1')
#jg2 = JobGroup(id=2, name='group2', first_job_id=5, desc='It is group2')
#job1 = Job(id=1, group_id=1, name='A', desc='jobA', running_timeout=60, trigger='0 * * * *')
#job2 = Job(id=2, group_id=1, name='B', desc='jobB', running_timeout=60, trigger='1 * * * *')
#job3 = Job(id=3, group_id=1, name='C', desc='jobC', running_timeout=60, trigger='2 * * * *')
#job4 = Job(id=4, group_id=1, name='D', desc='JobE', running_timeout=60, trigger='3 * * * *')
#job5 = Job(id=5, group_id=2, name='E', desc='jobF', running_timeout=60, trigger='4 * * * *')
#job6 = Job(id=6, group_id=2, name='F', desc='JobD', running_timeout=60, trigger='5 * * * *')
#job7 = Job(id=7, group_id=2, name='G', desc='jobG', running_timeout=60, trigger='6 * * * *')
#session.add(jg1)
#session.add(jg2)
#session.commit()
#session.add(job1)
#session.add(job2)
#session.add(job3)
#session.add(job4)
#session.add(job5)
#session.add(job6)
#session.add(job7)
#session.commit()
#session.close()

master.serve_forever(debug=True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


class Task(object):
    """
    Contains the properties of task, which is the instance of Job
    """

    #__slots__ = ("id", "job_id", "group", "group_id", "runtime", "worker_id", "running_timeout", "done_time", "state")

    STATE_INIT = 100
    STATE_TIMING = 101
    STATE_RUNNING = 102
    STATE_SCHEDULE_TIMEOUT = 103
    STATE_RUNNING_TIMEOUT = 104
    STATE_DOWNLOAD_FAIL = 105
    STATE_SUCCESS = 0
    STATE_FAIL = 1

    #INIT_STATE = 0
    #TIMING_STATE = 1
    #SUCCESS_STATE = 2
    #RUNNING_STATE = 3
    #SCHEDULE_TIMEOUT_STATE = 4
    #RUNNING_TIMEOUT_STATE = 5
    #FAIL = 6

    def __init__(self, id=None, job_id=None, group_id=None, runtime=None, running_timeout=None,
                 worker_id=None, done_time=None, state=STATE_INIT):
        self.id = id
        self.job_id = job_id
        self.group_id = group_id
        self.runtime = runtime
        self.running_timeout = running_timeout
        self.worker_id = worker_id
        self.done_time = done_time
        self.state = state

    def __str__(self):
        return "Task(id=%d, job_id=%d, runtime=%d)" % (self.id, self.job_id, self.runtime)

    def __repr__(self):
        return "Task(id=%d, job_id=%d, runtime=%d)" % (self.id, self.job_id, self.runtime)

    def __hash__(self):
        return (self.runtime & 65535) * self.job_id

    def __eq__(self, other):
        return isinstance(other, Task) and self.runtime == other.runtime and \
                self.job_id == other.job_id

    def to_dict(self):
        return {"id":self.id, "job_id":self.job_id, "group_id":self.group_id,
                "runtime":self.runtime, "worker_id":self.worker_id, "done_time":self.done_time,
                "state":self.state}

